
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

entity DirectCurrentControl is
	Port (
		-- Measure
 		meas_in : in std_logic_vector(15 downto 0) := (others => '0');
		-- Reference
		ref_in : in std_logic_vector(15 downto 0) := (others => '0');
		-- Tolerence
		tol_in  : in std_logic_vector(15 downto 0) := (others => '0');
		-- Delay
		delay_in  : in std_logic_vector(15 downto 0) := (others => '0');
		-- Error = Measure - Reference
		err_out  : out std_logic_vector(15 downto 0);

		-- Timing pulses
		adc_done_pulse_in : in std_logic;
		data_valid_pulse_in : in std_logic;

		-- PWM output
		pwm_out : out std_logic;

		-- Main clock running at 250 MHz
		clk_in : in std_logic
	);
end DirectCurrentControl;

architecture impl of DirectCurrentControl is

	ATTRIBUTE X_INTERFACE_INFO : STRING;
	ATTRIBUTE X_INTERFACE_INFO of clk_in: SIGNAL is "xilinx.com:signal:clock:1.0 clk CLK";

	signal meas_ff1 : std_logic_vector(15 downto 0) := (others => '0');
	signal ref_ff1 : std_logic_vector(15 downto 0) := (others => '0');
	signal ref_ff2 : std_logic_vector(15 downto 0) := (others => '0');
	signal tol_ff1 : std_logic_vector(15 downto 0) := (others => '0');
	signal tol_ff2 : std_logic_vector(15 downto 0) := (others => '0');

	signal counter_reg : unsigned(15 downto 0) := (others => '0');

	signal err : signed(15 downto 0);

	type type_fsm_state is (STATE_HIGH, STATE_LOW, STATE_H2L, STATE_L2H);
	signal fsm_state : type_fsm_state := STATE_LOW;

begin

	INPUT : process(clk_in)
	begin
		if rising_edge(clk_in) then

			-- new SBO data are available
			if data_valid_pulse_in = '1' then
				ref_ff1 <= ref_in;
				tol_ff1 <= tol_in;
			end if;

			-- new measure is available
			if adc_done_pulse_in = '1' then
				meas_ff1 <= meas_in;
				ref_ff2 <= ref_ff1;
				tol_ff2 <= tol_ff1;
			end if;

		end if;
	end process INPUT;

	err <= signed(meas_ff1) - signed(ref_ff2);
	err_out <= std_logic_vector(err);

	HYSTERESIS : process(clk_in)
	begin
		if rising_edge(clk_in) then

			case fsm_state is

				when STATE_HIGH =>
					pwm_out <= '0';
					counter_reg <= (others => '0');
					if err < -signed(tol_ff2) then
						fsm_state <= STATE_H2L;
					end if;

				when STATE_H2L =>
					pwm_out <= '1';
					counter_reg <= counter_reg + 1;
					if counter_reg >= unsigned(delay_in) then
						fsm_state <= STATE_LOW;
					end if;

				when STATE_LOW =>
					pwm_out <= '1';
					counter_reg <= (others => '0');
					if err > signed(tol_ff2) then
						fsm_state <= STATE_L2H;
					end if;

				when STATE_L2H =>
					pwm_out <= '0';
					counter_reg <= counter_reg + 1;
					if counter_reg >= unsigned(delay_in) then
						fsm_state <= STATE_HIGH;
					end if;

				when others => null;
	    end case;

	  end if;
	end process HYSTERESIS;

end impl;
